/***********************************
 * Author:        Matthew Cass     *
 * Email:         mcass@nist.gov   *
 * Date:          June 6, 2002     *
 *modified "Thu Mar 24 09:42:13 2011" *by "Paul E. Black"
 *                                 *
 * This is the implementation file *
 * for the density matrix class.   *
 **********************************/

#include <iostream>
using namespace std;
#include <iomanip>
#include <fstream>
#include <stdlib.h>	// for exit()
#include "denmat.h"
#include <math.h>

static string versionRCS =
"$Header: /home/black/Quantum/QCSim/RCS/denmat.C,v 1.9 2011/03/24 17:52:27 black Exp $";

string denmat$version() {
    return versionRCS;
}

//CONSTRUCTOR

denmat::denmat(int numBits)
{
    mat_size = (int)pow(2, numBits);
    int i;

    if (numBits < 1) {
	exit(1);
    }

    mat_bits = numBits;

    matrix = new float complex*[mat_size];

    for(i = 0; i < mat_size; i++) {
	matrix[i] = new float complex[mat_size];
    }
}

denmat::~denmat()
{
  int i;

  for (i = 0; i < mat_size; i++)
    delete [] matrix[i];
  delete [] matrix;
}

// check that the density matrix is well-formed
void denmat::check()
{
    double sumDiags = 0;

    if (mat_bits < 1 || mat_size != (int)pow(2, mat_bits)) {
	cout << "Size is " << mat_size << ", but that is not 2^" 
	     << mat_bits << " (number of bits)\n";
	return;
    }

    // SKIMP - correlations should be ||c|| < .5, I think

    // probabilities (diagonals) must be real and between 0 and 1
    for (int row = 0; row < mat_size; row++) {
	if (__imag__ this->get_state(row, row) != 0) {
	    cout << "Imaginary probability " << this->get_state(row, row)
				<< " at (" << row << ", " << row << ")\n";
	    return;
	}
	double prob = __real__ this->get_state(row, row);
	if (prob < 0 || prob > 1) {
	    cout << "Invalid probability " << prob
				<< " at (" << row << ", " << row << ")\n";
	    return;
	}
	sumDiags += prob;
    }

    // probabilities must sum to (nearly) 1
    if (fabs(sumDiags-1) > 1e-07) {
	cout << "Probabilities differ from 1 by " << (sumDiags-1.) << "\n";
	return;
    }
}

denmat::denmat *init_denmat(denmat *d, float complex ket[])
{
    int i, j;
    int size = d->size();

    // fill with 0
    for (i = 0; i < size; i++) {
	for (j = 0; j < size; j++) {
	    d->matrix[i][j] = 0;
	}
    }

    // put cross products of amplitudes in density matrix
    for (i = 0; i < size; i++) {
	for (j = 0; j < size; j++) {
	    // SKIMP: no check that the array is the right size
	    d->set_state(i, j, ket[i] * conj(ket[j]));
	}
    }

    return d;
}

void denmat::print_state()
{
    //ofstream fout;
    int i, j;
  
    //fout.open("output");  
    for (i = 0; i < mat_size; i++) {
	for (j = 0; j < mat_size; j++) {
	    cout << " ";
	    if (__imag__ matrix[i][j] == 0) {
		cout << setw(7) << __real__ matrix[i][j];
	    } else {
		cout << setw(2) << __real__ matrix[i][j];
		if (__imag__ matrix[i][j] > 0) {
		    cout << "+";
		}
		cout << __imag__ matrix[i][j] << "i";
	    }
	}
	cout << endl;
    }
    //fout.close();
}    
        
void denmat::set_state(int r, int c, float complex v)
{
    matrix[r][c] = v;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/denmat.C,v $
