// $Id: opOneQubit.h,v 1.6 2003/06/13 16:22:52 black Exp $
// *created  "Thu Sep 19 14:24:02 2002" *by "Paul E. Black"
// *modified "Fri Jun 13 11:24:00 2003" *by "Paul E. Black"

// Use Cass' observation as a template (macro) for a function to do a 
// general, single qubit operation on a density matrix.

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of S
// and Technology by employees of the Federal Government in t
// of their official duties.  Pursuant to title 17 Section 10
// United States Code this software is not subject to copyrig
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

// Matthew Cass found (August 2002) that for a Hadamard, quadruples of values,
// at the corners of a 2^t square (where t is numberOfQubits - target qubit),
// are mixed together.  In the following a and b are 2^t columns apart,
// and a and c are 2^t rows apart.  In the code below the variables are 
// ul (upper, left), ur, ll, and lr.
//
//	{ ....  a .... b ..... }
//	{       .      .       }
//	{       .      .       }
//	{       .      .       }
//	{ ....  c .... d ..... }
// becomes
//	{ . a+b+c+d  a-b+c-d . }
//	{       .      .       }
//	{       .      .       }
//	{       .      .       }
//	{ . a+b-c-d  a-b-c+d . }

// We can generalize Cass' observation by asking what a general operation
//	{ e f }
//	{ g h }
// does to a density matrix, that is, what is the result of
//	{ e f } . { ul ur } . { e' g' }
//	{ g h }   { ll lr }   { f' h' }
// where ' is the conjugate.  The result is
//	{ ee'ul+ef'ur+fe'll+ff'lr eg'ul+eh'ur+fg'll+fh'lr }
//	{ ge'ul+gf'ur+he'll+hf'lr gg'ul+gh'ur+hg'll+hh'lr }
// Substitute the definition of a Hadamard
//	{ 1  1 }
//	{ 1 -1 }
// to verify the formula gives the same result as the specific Hadamard
// worked out above.  More importantly, we can apply Cass' observation
// (and verify by tensor products with Identity) to see that these general
// operations are used over and over locally across the density matrix
// in blocks.

// A block is a 2^(t+1) square, suggested here by a0 (upper, lefthand 
// corner) to b2^(t-1) (upper, righthand corner) to y0 (lower, lefthand 
// corner) to z2^(t-1) (lower, righthand corner).  As above a0, b0, c0, 
// and d0 are mixed, as are a1, b1, c1, d1, and e0, f0, g0, h0, down to
// w2^(t-1), x2^(t-1), y2^(t-1), and z2^(t-1).
//
//	{ a0 a1 .. a2^(t-1) b0 b1 .. b2^(t-1) ..... }
//	{ e0 e1 .. e2^(t-1) f0 f1 .. f2^(t-1) ..... }
//	{   .                 .                 .   }
//	{   .                 .                 .   }
//	{ w0 w1 .. w2^(t-1) x0 x1 .. x2^(t-1) ..... }
//	{ c0 c1 .. c2^(t-1) d0 d1 .. d2^(t-1) ..... }
//	{ g0 g1 .. g2^(t-1) h0 h1 .. h2^(t-1) ..... }
//	{   .                 .                 .   }
//	{   .                 .                 .   }
//	{ y0 y1 .. y2^(t-1) z0 z1 .. z2^(t-1) ..... }
//	{   .                 .                 .   }
//	{   .                 .                 .   }
//	{   .                 .                 .   }
//
// Blocks are repeated at 2^(t+1) intervals across and down the density 
// matrix.  So the structure of the code is two loops to do all the blocks, 
// and two loops to do the rows and columns in each block.  The loops are 
// switched and interleaved to work well with the array indexing, memory 
// allocation, and paging schemes.

// An operation on a single qubit can be built by simply specifying
// what of ul, ur, ll, and lr goes in each final place.  This macro
// takes care of looping over and within blocks, temporaries, etc.

#define PARAM_COMMA , // to put commas in a parameter list
#define OprOneQubit(FNNAME, PARAMETERS, DECLARATIONS, INNER_OPERATION) \
void FNNAME(int targetBit, denmat *mat PARAMETERS)\
{\
    int tbit = (mat->bits()-1) - targetBit;\
    int target = (int)pow(2, tbit);\
    int nextblock = 2 * target;\
\
    /* global declarations, e.g., temporary storage, intermediate values */\
    DECLARATIONS\
\
    /* do every row of blocks */\
    for (int rowblock = 0; rowblock < mat->size(); rowblock += nextblock) {\
	/* do every row in that block */\
	for (int row = rowblock; row < rowblock+target; row++) {\
	    /* do every column of blocks */\
	    for (int colblock = 0; colblock < mat->size(); colblock += nextblock) {\
		/* do every column in that block */\
		for (int col = colblock; col < colblock+target; col++) {\
		    /* do the four corners of the square.  row, col is the */\
		    /* upper, leftmost entry */\
/* uncomment this stuff if you want to check array accesses\
	be sure to #include <iostream>\
		    if (row < 0 || row >= mat->size()) {\
			cout << "invalid row: " << row << endl;\
			return;\
		    }\
		    if (row+target < 0 || row+target >= mat->size()) {\
			cout << "invalid row+target: " << row+target << endl;\
			return;\
		    }\
		    if (col < 0 || col >= mat->size()) {\
			cout << "invalid col: " << col << endl;\
			return;\
		    }\
		    if (col+target < 0 || col+target >= mat->size()) {\
			cout << "invalid col+target: " << col+target << endl;\
			return;\
		    }\
*/\
		    /* put results back */\
		    INNER_OPERATION\
		}\
	    }\
	}\
    }\
}

// end of $Source: /home/black/Quantum/QCSim/RCS/opOneQubit.h,v $
