// *created  "Thu Aug 15 13:01:00 2002" *by "Paul E. Black"
// *modified "Thu Mar 24 09:39:05 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

#include <math.h>
#include "operations.h"
#include "parsingFuncs.h"

static string versionRCS =
"$Header: /home/black/Quantum/QCSim/RCS/operations.C,v 1.30 2011/03/25 13:56:20 black Exp $";

string operations$version() {
    return versionRCS;
}

// flip the target bit where the control mask is true in the passed matrix
// this changes the passed matrix
//
// Matthew Cass found (July 2002) that CNOT swaps pairs of rows and pairs 
// of columns.  The loops are switched to work well with the array
// indexing, memory allocation, and paging schemes.
static void cnot_denmat(denmat *mat, int tbit, int control)
{
    int target  = (int)pow(2, tbit);
    int mask = control | target;
    float complex temp;

    // swap appropriate rows
    for (int row = 0; row < mat->size(); row++) {
	if ((mask & row) == mask) { 
	    // control and target bits are 1 in this row; swap with the 
	    // corresponding row where the target bit is zero
	    int otherrow = row ^ target;
	    for (int col = 0; col < mat->size(); col++) {
		temp = mat->get_state(row, col);
		mat->set_state(row, col, mat->get_state(otherrow, col));
		mat->set_state(otherrow, col, temp);
	    }
	}
    }
  
    // swap appropriate columns
    // NOTE: SWAPPING COLUMNS TOOK 4 TIMES AS LONG AS SWAPPING ROWS, 
    // PRESUMABLY BECAUSE OF CACHING.  THIS ROW-THEN-COLUMN LOOPING IS
    // THREE TIMES FASTER THAN CODING STRICTLY SIMILAR TO ROW SWAPPING.
    for (int row = 0; row < mat->size(); row++) {
	for (int col = 0; col < mat->size(); col++) {
	    if ((mask & col) == mask) { 
		// control and target bits are 1 in this column; swap this with
		// the corresponding column where the target bit is zero
		int othercol = col ^ target;
		// THE "for (row ..." LOOP WOULD BE HERE IF WRITTEN
		// STRICTLY SIMILAR TO THE ROW SWAPPING ABOVE.
		temp = mat->get_state(row, col);
		mat->set_state(row, col, mat->get_state(row, othercol));
		mat->set_state(row, othercol, temp);
	    }
	}
    }
}

// compute the mask for CNOT
ParmErrorType cnot(simstate *ss, int targetBit, int controlBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit < 0 || controlBit >= ss->Qubits()) {
	return QubitOOR;
    } else if (targetBit == controlBit) {
	return DupQubit;
    }

    // external order is qubit 0 first, qubit 1 second, etc.
    int tbit = ss->Qubits() - (targetBit+1);
    int cbit = ss->Qubits() - (controlBit+1);


    cnot_denmat(ss->get_denmat(), tbit, (int)pow(2, cbit));
    return NoParmError;
}

// compute the mask for Toffoli or CCNOT
ParmErrorType toffoli(simstate *ss, int targetBit, int controlBit1, int controlBit2)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       controlBit1 == controlBit2) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2));
    return NoParmError;
}


// compute the mask for C3NOT
ParmErrorType c3not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || controlBit1 == controlBit2 ||
	       controlBit1 == controlBit3 || controlBit2 == controlBit3) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3));
    return NoParmError;
}

// compute the mask for C4NOT
ParmErrorType c4not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit2 == controlBit3 ||
	       controlBit2 == controlBit4 || controlBit3 == controlBit4) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4));
    return NoParmError;
}

// compute the mask for C5NOT
ParmErrorType c5not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4, int controlBit5)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ||
	controlBit5 < 0 || controlBit5 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       targetBit == controlBit5 ||
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit1 == controlBit5 ||
	       controlBit2 == controlBit3 || controlBit2 == controlBit4 ||
	       controlBit2 == controlBit5 || controlBit3 == controlBit4 ||
	       controlBit3 == controlBit5 || controlBit4 == controlBit5) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);
    int cbit5 = ss->Qubits() - (controlBit5+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4) | (int)pow(2, cbit5));
    return NoParmError;
}

// compute the mask for C6NOT
ParmErrorType c6not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4, int controlBit5, int controlBit6)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ||
	controlBit5 < 0 || controlBit5 >= ss->Qubits() ||
	controlBit6 < 0 || controlBit6 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       targetBit == controlBit5 || targetBit == controlBit6 || 
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit1 == controlBit5 ||
	       controlBit1 == controlBit6 || controlBit2 == controlBit3 ||
	       controlBit2 == controlBit4 || controlBit2 == controlBit5 ||
	       controlBit2 == controlBit6 || controlBit3 == controlBit4 ||
	       controlBit3 == controlBit5 || controlBit3 == controlBit6 ||
	       controlBit4 == controlBit5 || controlBit4 == controlBit6 ||
	       controlBit5 == controlBit6) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);
    int cbit5 = ss->Qubits() - (controlBit5+1);
    int cbit6 = ss->Qubits() - (controlBit6+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4) | (int)pow(2, cbit5) | (int)pow(2, cbit6));
    return NoParmError;
}

// compute the mask for C7NOT
ParmErrorType c7not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4, int controlBit5, int controlBit6, int controlBit7)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ||
	controlBit5 < 0 || controlBit5 >= ss->Qubits() ||
	controlBit6 < 0 || controlBit6 >= ss->Qubits() ||
	controlBit7 < 0 || controlBit7 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       targetBit == controlBit5 || targetBit == controlBit6 || 
	       targetBit == controlBit7 ||
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit1 == controlBit5 ||
	       controlBit1 == controlBit6 || controlBit1 == controlBit7 ||
	       controlBit2 == controlBit3 || controlBit2 == controlBit4 ||
	       controlBit2 == controlBit5 || controlBit2 == controlBit6 ||
	       controlBit2 == controlBit7 || controlBit3 == controlBit4 ||
	       controlBit3 == controlBit5 || controlBit3 == controlBit6 ||
	       controlBit3 == controlBit7 || controlBit4 == controlBit5 ||
	       controlBit4 == controlBit6 || controlBit4 == controlBit7 ||
	       controlBit5 == controlBit6 || controlBit5 == controlBit7 ||
	       controlBit6 == controlBit7) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);
    int cbit5 = ss->Qubits() - (controlBit5+1);
    int cbit6 = ss->Qubits() - (controlBit6+1);
    int cbit7 = ss->Qubits() - (controlBit7+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4) | (int)pow(2, cbit5) | (int)pow(2, cbit6) | (int)pow(2, cbit7));
    return NoParmError;
}

// compute the mask for C8NOT
ParmErrorType c8not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4, int controlBit5, int controlBit6, int controlBit7, int controlBit8)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ||
	controlBit5 < 0 || controlBit5 >= ss->Qubits() ||
	controlBit6 < 0 || controlBit6 >= ss->Qubits() ||
	controlBit7 < 0 || controlBit7 >= ss->Qubits() ||
	controlBit8 < 0 || controlBit8 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       targetBit == controlBit5 || targetBit == controlBit6 || 
	       targetBit == controlBit7 || targetBit == controlBit8 ||
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit1 == controlBit5 ||
	       controlBit1 == controlBit6 || controlBit1 == controlBit7 ||
	       controlBit1 == controlBit8 || controlBit2 == controlBit3 ||
	       controlBit2 == controlBit4 || controlBit2 == controlBit5 ||
	       controlBit2 == controlBit6 || controlBit2 == controlBit7 ||
	       controlBit2 == controlBit8 || controlBit3 == controlBit4 ||
	       controlBit3 == controlBit5 || controlBit3 == controlBit6 ||
	       controlBit3 == controlBit7 || controlBit3 == controlBit8 ||
	       controlBit4 == controlBit5 || controlBit4 == controlBit6 ||
	       controlBit4 == controlBit7 || controlBit4 == controlBit8 ||
	       controlBit5 == controlBit6 || controlBit5 == controlBit7 ||
	       controlBit5 == controlBit8 || controlBit6 == controlBit7 ||
	       controlBit6 == controlBit8 || controlBit7 == controlBit8) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);
    int cbit5 = ss->Qubits() - (controlBit5+1);
    int cbit6 = ss->Qubits() - (controlBit6+1);
    int cbit7 = ss->Qubits() - (controlBit7+1);
    int cbit8 = ss->Qubits() - (controlBit8+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4) | (int)pow(2, cbit5) | (int)pow(2, cbit6) | (int)pow(2, cbit7) | (int)pow(2, cbit8));
    return NoParmError;
}

// compute the mask for C9NOT
ParmErrorType c9not(simstate *ss, int targetBit, int controlBit1, int controlBit2, int controlBit3, int controlBit4, int controlBit5, int controlBit6, int controlBit7, int controlBit8, int controlBit9)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit1 < 0 || controlBit1 >= ss->Qubits() ||
	controlBit2 < 0 || controlBit2 >= ss->Qubits() ||
	controlBit3 < 0 || controlBit3 >= ss->Qubits() ||
	controlBit4 < 0 || controlBit4 >= ss->Qubits() ||
	controlBit5 < 0 || controlBit5 >= ss->Qubits() ||
	controlBit6 < 0 || controlBit6 >= ss->Qubits() ||
	controlBit7 < 0 || controlBit7 >= ss->Qubits() ||
	controlBit8 < 0 || controlBit8 >= ss->Qubits() ||
	controlBit9 < 0 || controlBit9 >= ss->Qubits() ){
	return QubitOOR;
    } else if (targetBit == controlBit1 || targetBit == controlBit2 ||
	       targetBit == controlBit3 || targetBit == controlBit4 ||
	       targetBit == controlBit5 || targetBit == controlBit6 || 
	       targetBit == controlBit7 || targetBit == controlBit8 ||
	       targetBit == controlBit9 ||
	       controlBit1 == controlBit2 || controlBit1 == controlBit3 || 
	       controlBit1 == controlBit4 || controlBit1 == controlBit5 ||
	       controlBit1 == controlBit6 || controlBit1 == controlBit7 ||
	       controlBit1 == controlBit8 || controlBit1 == controlBit9 ||
	       controlBit2 == controlBit3 || controlBit2 == controlBit4 ||
	       controlBit2 == controlBit5 || controlBit2 == controlBit6 ||
	       controlBit2 == controlBit7 || controlBit2 == controlBit8 ||
	       controlBit2 == controlBit9 || controlBit3 == controlBit4 ||
	       controlBit3 == controlBit5 || controlBit3 == controlBit6 ||
	       controlBit3 == controlBit7 || controlBit3 == controlBit8 ||
	       controlBit3 == controlBit9 || controlBit4 == controlBit5 ||
	       controlBit4 == controlBit6 || controlBit4 == controlBit7 ||
	       controlBit4 == controlBit8 || controlBit4 == controlBit9 || 
	       controlBit5 == controlBit6 || controlBit5 == controlBit7 ||
	       controlBit5 == controlBit8 || controlBit5 == controlBit9 ||
	       controlBit6 == controlBit7 || controlBit6 == controlBit8 ||
	       controlBit6 == controlBit9 || controlBit7 == controlBit8 ||
	       controlBit7 == controlBit9 || controlBit8 == controlBit9) {
	return DupQubit;
    }

    int tbit  = ss->Qubits() - (targetBit+1);
    int cbit1 = ss->Qubits() - (controlBit1+1);
    int cbit2 = ss->Qubits() - (controlBit2+1);
    int cbit3 = ss->Qubits() - (controlBit3+1);
    int cbit4 = ss->Qubits() - (controlBit4+1);
    int cbit5 = ss->Qubits() - (controlBit5+1);
    int cbit6 = ss->Qubits() - (controlBit6+1);
    int cbit7 = ss->Qubits() - (controlBit7+1);
    int cbit8 = ss->Qubits() - (controlBit8+1);
    int cbit9 = ss->Qubits() - (controlBit9+1);

    cnot_denmat(ss->get_denmat(), tbit,
		(int)pow(2, cbit1) | (int)pow(2, cbit2) | (int)pow(2, cbit3) | (int)pow(2, cbit4) | (int)pow(2, cbit5) | (int)pow(2, cbit6) | (int)pow(2, cbit7) | (int)pow(2, cbit8) | (int)pow(2, cbit9));
    return NoParmError;
}

// do controlled Z on targetBit in passed matrix
// this changes the passed matrix
ParmErrorType copZ(simstate *ss, int targetBit, int controlBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit < 0 || controlBit >= ss->Qubits()) {
	return QubitOOR;
    } else if (targetBit == controlBit) {
	return DupQubit;
    }

    // external order is qubit 0 first, qubit 1 second, etc.
    int tbit = ss->Qubits() - (targetBit+1);
    int cbit = ss->Qubits() - (controlBit+1);
    int mask = (int)pow(2, tbit) | (int)pow(2, cbit);
    denmat *mat = ss->get_denmat();

    // negate appropriate rows
    for (int row = 0; row < mat->size(); row++) {
	if ((mask & row) == mask) { 
	    // control and target bits are 1 in this row;
	    // negate all values in this column
	    for (int col = 0; col < mat->size(); col++) {
		mat->set_state(row, col, 0-mat->get_state(row, col));
	    }
	}
    }
  
    // negate appropriate columns
    // NOTE: THIS ROW-THEN-COLUMN LOOPING IS TWICE AS FAST AS
    // CODING STRICTLY SIMILAR TO THE ABOVE.
    for (int row = 0; row < mat->size(); row++) {
	for (int col = 0; col < mat->size(); col++) {
	    if ((mask & col) == mask) { 
		// control and target bits are 1 in this column;
		// negate all values in this row
		// THE "for (row ..." LOOP WOULD BE HERE IF WRITTEN
		// STRICTLY SIMILAR TO THE ROW SWAPPING ABOVE.
		mat->set_state(row, col, 0-mat->get_state(row, col));
	    }
	}
    }
        return NoParmError;
}

// macro to define operations on one qubit of a density matrix
#include "opOneQubit.h"

// perform Hadamard on targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(hadamard_denmat,
	/* no extra parameters needed */
,
	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);
	/* put results back */
	mat->set_state(row,        col,
			(ul + ur + ll + lr)/2.F);
	mat->set_state(row,        col+target,
			(ul - ur + ll - lr)/2.F);
	mat->set_state(row+target, col,
			(ul + ur - ll - lr)/2.F);
	mat->set_state(row+target, col+target,
			(ul - ur - ll + lr)/2.F);
)

// wrapper - simulation state is a density matrix
ParmErrorType hadamard(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits()) {
	return QubitOOR;
    }

    hadamard_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}

// rotate qubit 2theta radians about Y axis
static OprOneQubit(rotateY_denmat,
	PARAM_COMMA float theta
,
	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
	float complex newul;
	float complex newur;
	double sintheta = sin(theta);
	double costheta = cos(theta);
	/* float (instead of double) for cygwin portability */
	float cos2   = costheta * costheta;
	float cossin = costheta * sintheta;
	float sin2   = 1 - cos2;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);
	/* put results back */
	newul = ul*cos2 + (ur + ll)*cossin + lr*sin2;
	mat->set_state(row,        col,
			newul);
	newur = ur*cos2 + (lr - ul)*cossin - ll*sin2;
	mat->set_state(row,        col+target,
			newur);
	mat->set_state(row+target, col,
			ll - ur + newur);
	mat->set_state(row+target, col+target,
			ul + lr - newul);
)

// from CRC Standard Mathematical Tables, 14th ed., pg 16, 1965
static const float PI = 3.141592653589793238;

ParmErrorType rotateY(simstate *ss, float angle, int qubit)
{
    if (angle < -2*PI || angle > 2*PI) {
	return ProbOOR;
    } else if (qubit < 0 || qubit >= ss->Qubits()) {
	return QubitOOR;
    }

    //ss->get_denmat()->check();

    rotateY_denmat(qubit, ss->get_denmat(), angle/2.);

    //ss->get_denmat()->check();

    return NoParmError;
}

// perform bit flip or X on targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(opX_denmat,
	/* no extra parameters needed */
,
	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);

	/* put results back */
	mat->set_state(row,        col,		lr);
	mat->set_state(row,        col+target,	ll);
	mat->set_state(row+target, col,		ur);
	mat->set_state(row+target, col+target,  ul);
)

// wrapper - simulation state is a density matrix
ParmErrorType opX(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits()) {
	return QubitOOR;
    }

    opX_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}

// perform pauli-Y on targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(opY_denmat,
	/* no extra parameters needed */
,
	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);

	/* put results back */
	mat->set_state(row,        col,		 lr);
	mat->set_state(row,        col+target,	0-ll);
	mat->set_state(row+target, col,		0-ur);
	mat->set_state(row+target, col+target,   ul);
)

// wrapper - simulation state is a density matrix
ParmErrorType opY(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits()) {
	return QubitOOR;
    }

    opY_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}


// perform pauli-Z on targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(opZ_denmat,
	/* no extra parameters needed */
,
	/* no temporary storage needed */
,
	/* negate two correlations */
	mat->set_state(row, col+target,	0-mat->get_state(row, col+target));
	mat->set_state(row+target, col, 0-mat->get_state(row+target,col));
)

// wrapper - simulation state is a density matrix
ParmErrorType opZ(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits()) {
	return QubitOOR;
    }

    opZ_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}

// perform a probabilistic XYZ flip on targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(probXYZ_denmat,
		   PARAM_COMMA float p
		   PARAM_COMMA float q
		   PARAM_COMMA float r
,
	float one_minus_pq = 1 - p - q;
	float p_plus_q = p + q;
	float p_minus_q = p - q;
	float one_minus_pq2r = 1 - p - q - 2*r;

	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);

	/* put results back */
	mat->set_state(row, col,	one_minus_pq*ul + p_plus_q*lr);
	mat->set_state(row, col+target,	one_minus_pq2r*ur + p_minus_q*ll);
	mat->set_state(row+target, col,	one_minus_pq2r*ll + p_minus_q*ur);
	mat->set_state(row+target, col+target, one_minus_pq*lr + p_plus_q*ul);
)

// wrapper - simulation state is a density matrix
ParmErrorType probabilisticXYZ(simstate *ss, float probP, float probQ, float probR, int targetBit)
{
    float probSum;
    probSum = probP + probQ + probR;

    if (targetBit < 0 || targetBit >= ss->Qubits() ){
	return QubitOOR;
    } else if (probP < 0.0 || probP > 1.0) {
	return ProbOOR;
    } else if (probQ < 0.0 || probQ > 1.0) {
	return ProbOOR;
    } else if (probR < 0.0 || probR > 1.0) {
	return ProbOOR;
    } else if (fabs(probSum - 1.0) < .00000000001) {
	// equal - do nothing
    } else if (float(probSum - 1.0) > float(0.0)) {
	return ProbOOR;
    }

    probXYZ_denmat(targetBit, ss->get_denmat(), probP, probQ, probR);
    return NoParmError;
}

// Generalized Amplitude Damping
// see Nielsen & Chuang Sect. 8.3.5 page 382
// probDamp is the probability of damping - gamma in N&C
//	0 is no damping (I), 1 is force to p|0>+(1-p)|1>
// p is the probability of going to zero
static OprOneQubit(genAmpDamp_denmat,
		   PARAM_COMMA float probDamp
		   PARAM_COMMA float p
,
	float sqrt_one_minus_gamma = sqrt(1. - probDamp);

	/* temporary storage for each quadruple of values */
	float complex a;
	float complex b;
	float complex c;
	float complex d;
,
	/* get the quadruple of values from the array */
	a = mat->get_state(row,       col);
	b = mat->get_state(row,       col+target);
	c = mat->get_state(row+target,col);
	d = mat->get_state(row+target,col+target);

	/* put results back */
	mat->set_state(row,	   col,	       a + probDamp*(p*d - (1-p)*a));
	mat->set_state(row,	   col+target, b*sqrt_one_minus_gamma);
	mat->set_state(row+target, col,	       c*sqrt_one_minus_gamma);
	mat->set_state(row+target, col+target, d + probDamp*((1-p)*a - p*d));
)

ParmErrorType genAmpDamp(simstate *ss, float probDamp, float probZero, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ){
	return QubitOOR;
    } else if (probDamp < 0.0 || probDamp > 1.0) {
	return ProbOOR;
    } else if (probZero < 0.0 || probZero > 1.0) {
	return ProbOOR;
    }

    genAmpDamp_denmat(targetBit, ss->get_denmat(), probDamp, probZero);
    return NoParmError;
}

// perform a noisy Hadamard on targetBit in passed matrix
// this models Eric Hsu's synthetic Hadamard
// this changes the passed matrix
static OprOneQubit(noisyHadamard_denmat,
	/* no extra parameters needed */
,
	/* temporary storage for each quadruple of values */
	float complex ul;
	float complex ur;
	float complex ll;
	float complex lr;
,
	/* get the quadruple of values from the array */
	ul = mat->get_state(row,       col);
	ur = mat->get_state(row,       col+target);
	ll = mat->get_state(row+target,col);
	lr = mat->get_state(row+target,col+target);

	/* put results back */
	mat->set_state(row,        col,
		 .405772F*ll + .492518F*lr + .495482F*ul + .405772F*ur);
	mat->set_state(row,        col+target,
		 .408228F*ll - .408228F*lr + .408228F*ul - .405246F*ur);
	mat->set_state(row+target, col,
		-.405246F*ll - .408228F*lr + .408228F*ul + .408228F*ur);
	mat->set_state(row+target, col+target,
		-.405772F*ll + .507482F*lr + .504518F*ul - .405772F*ur);
)

// compute the result of a noisy Hadamard
ParmErrorType noisyHadamard_01(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ) {
	return QubitOOR;
    }

    noisyHadamard_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}

// measure (collapse to a classical state) targetBit in passed matrix
// this changes the passed matrix
static OprOneQubit(measure_denmat,
	/* no extra parameters needed */
,
	/* no globals declarations needed */
,
	/* zero the correlations */
	mat->set_state(row,        col+target, 0);
	mat->set_state(row+target, col,        0);
)

// wrapper - simulation state is a density matrix
ParmErrorType measure(simstate *ss, int targetBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ) {
	return QubitOOR;
    }

    measure_denmat(targetBit, ss->get_denmat());
    return NoParmError;
}

static void chadamard_denmat(denmat *mat, int tbit, int control)
{
    int target  = (int)pow(2, tbit);
    int mask = control | target;
    float complex temp;
    float complex temp2;
    float sqrt2 = sqrt(2.F);

    // perform operations on rows
    for (int row = 0; row < mat->size(); row++) {
	if ((mask & row) == mask) { 
	    // control and target bits are 1 in this row
	    int otherrow = row ^ target;
	    for (int col = 0; col < mat->size(); col++) {
		temp = mat->get_state(row, col);
		temp2 = mat->get_state(otherrow, col);
		mat->set_state(row, col, (temp2 - temp)/sqrt2);
		mat->set_state(otherrow, col, (temp2 + temp)/sqrt2);
	    }
	}
    }
  
    // perform operations on columns
    for (int row = 0; row < mat->size(); row++) {
	for (int col = 0; col < mat->size(); col++) {
	    if ((mask & col) == mask) { 
		// control and target bits are 1 in this column
		int othercol = col ^ target;

		temp = mat->get_state(row, col);
		temp2 = mat->get_state(row, othercol);

		mat->set_state(row, col, (temp2 - temp)/sqrt2);
		mat->set_state(row, othercol, (temp2 + temp)/sqrt2);

	    }
	}
    }
}

// compute the mask for chadamard
ParmErrorType chadamard(simstate *ss, int targetBit, int controlBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ||
	controlBit < 0 || controlBit >= ss->Qubits()) {
	return QubitOOR;
    } else if (targetBit == controlBit) {
	return DupQubit;
    }

    // external order is qubit 0 first, qubit 1 second, etc.
    int tbit = ss->Qubits() - (targetBit+1);
    int cbit = ss->Qubits() - (controlBit+1);

    chadamard_denmat(ss->get_denmat(), tbit, (int)pow(2, cbit));
    return NoParmError;
}

// depolarizing error
// rho' = p*rho + (1-p)*Tr[rho]x1/4
static void BSWDepolar_denmat(denmat *mat, int fastestBit, float p)
{
    int minbit = (mat->bits() - 1) - fastestBit;
    int stride = (int) pow(2, minbit);
    int nextblock = 4 * stride;
    float one_minus_p_over_4 = (1 - p) / 4.;
    float complex aveDiag; // (1-p)(r00c00 + r01c01 + r10c10 + r11c11)/4

    // cout << "Minimum bit " << "(" << minbit << ")\n";
    // cout << "Stride " << stride << "\n";

    for (int rowblock = 0; rowblock < mat->size(); rowblock += nextblock) {
	for (int row = rowblock; row < rowblock + stride; row++) {
	    aveDiag = one_minus_p_over_4 * (
			mat->get_state(row, row) + 
			mat->get_state(row + stride, row + stride) +
			mat->get_state(row+2*stride, row+2*stride) +
			mat->get_state(row+3*stride, row+3*stride));

	    for (int colblock = 0; colblock < mat->size();
		 colblock += nextblock) {
		for (int col = colblock; col < colblock + stride; col++) {
		    // cout << "Starting submatrix at "<<row<<", "<<col<<"\n";
		    // do each entry of the (sub)matrix
		    for (int r = row; r < row+4*stride; r += stride) {
			for (int c = col; c < col+4*stride; c += stride) {
			    if (r == c) {
				mat->set_state(r, c, aveDiag +
						p * mat->get_state(r, c));
			    } else {
				mat->set_state(r, c,
						p * mat->get_state(r, c));
			    }
			}
		    }
		}
	    }
	}
    }
}

ParmErrorType BSWDepolCPhase(simstate *ss, float probSuccess, int targetBit, int controlBit)
{
    if (targetBit < 0 || targetBit >= ss->Qubits() ){
	return QubitOOR;
    } else if (controlBit < 0 || controlBit >= ss->Qubits() ){
	return QubitOOR;
    } else if (probSuccess < 0.0 || probSuccess > 1.0) {
	return ProbOOR;
    }

    // do the controlled Z
    ParmErrorType err = copZ(ss, targetBit, controlBit);
    if (err != NoParmError) return err;

    // Fortunately controlled Z doesn't change the diagonals, so we
    // can do the depolarization as a second step
    int fastestQubit = targetBit;
    if (controlBit > targetBit) fastestQubit = controlBit;
    BSWDepolar_denmat(ss->get_denmat(), fastestQubit, probSuccess);

    return NoParmError;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/operations.C,v $
