// *created  "Tue Jun 10 14:19:28 2003" *by "Paul E. Black"
// *modified "Thu Mar 24 10:35:18 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	Grammar and functions to parse jaQuzzi input
//-----------------------------------------------------------------------------

#include <stdio.h>     // for sprintf
#include <string.h>
#include "parsingFuncs.h"
#include "parseJAQ.h"

static string versionRCS = 
"$Header: /home/black/Quantum/QCSim/RCS/parseJAQ.C,v 1.7 2011/03/25 13:59:15 black Exp $";
string parseJAQ$version()
{
    return versionRCS;
}

//
//	VARIABLE DECLARATION STATEMENTS
//

extern int varNameLookup(char *name); // look up in symbol table
extern bool addToSymbolTab(char *name, int qubitNumb);
extern void nextTypeInfo();

extern void getSymbolByIndex(int ind, char* returnBuf);


// reserved keywords
static const char *keywords[] = { "variable", "qubit" };
static int NUM_KEYWORDS = sizeof(keywords) / sizeof(*keywords);

static GateBitType gateBitsArray[35];

static int numberOfVariables = 0;

static bool isReservedWord(char *buf)
{
    // check for reserved keywords
    for (int i = 0; i < NUM_KEYWORDS; i++) {
	if (strcmp(keywords[i], buf) == 0) {
	    return true;
	}
    }

    return false;		// if we get here everything is ok
}


static void parseBitOp(int ind)
{
    GateBitType bit = ErrorBit; // init as error

    if (nextType == NameSymbType) {
	if (strcmp(symbName, "H") == 0) {
	    bit = HadBit;
	} else if (strcmp(symbName, "NOT") == 0) {
	    bit = NotBit;
	} else if (strcmp(symbName, "sigma_y") == 0) {
	    bit = YBit;
	} else if (strcmp(symbName, "sigma_z") == 0) {
	    bit = ZBit;
	} else if (strcmp(symbName, "u") == 0) {
	    bit = Reset0Bit;
	} else if (strcmp(symbName, "d") == 0) {
	    bit = Reset1Bit;	
	}
    } else if (nextType == NumberSymbType) {
	if (symbNumber == 1) {
	    bit = ControlBit;
	}
    } else if (nextType == PuncSymbType) {
	if (symbPunct == '!') {
	    bit = MeasureBit;
	} else if (symbPunct == '-') {
	    bit = BlankBit;
	}
    }

    if (bit == ErrorBit) {
	dieWithError("Unknown bit in gate.");
    }

    gateBitsArray[ind] = bit;
}

static void processMeasureGate(int target)
{
    char symbolBuf[50];

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] != BlankBit && i != target) {
	    dieWithError("Unknown gate.");
	}
    }
    getSymbolByIndex(target, symbolBuf);
    cout << "measure(" << symbolBuf << ");\n";
}

static void processReset0Gate(int target)
{
    char symbolBuf[50];

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] != BlankBit && i != target) {
	    dieWithError("Unknown gate.");
	}
    }
    getSymbolByIndex(target, symbolBuf);
    cout << "genAmpDamp(1.0, 1.0, " << symbolBuf << ");\n";
}

static void processReset1Gate(int target)
{
    char symbolBuf[50];

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] != BlankBit && i != target) {
	    dieWithError("Unknown gate.");
	}
    }
    getSymbolByIndex(target, symbolBuf);
    cout << "genAmpDamp(1.0, 0.0, " << symbolBuf << ");\n";
}

static void processNotGate(int target)
{
    char symbolBuf[50];
    int controlBit1 = -1;
    int controlBit2 = -1;
    int controlBit3 = -1;
    int controlBit4 = -1;
    int controlBit5 = -1;
    int controlBit6 = -1;
    int controlBit7 = -1;
    int controlBit8 = -1;
    int controlBit9 = -1;
    int counter=0;

    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] == ControlBit) {
	    if (controlBit1 == -1) {
		controlBit1 = i;
	    } else if (controlBit2 == -1) {
		controlBit2 = i;
	    } else if (controlBit3 == -1) {
		controlBit3 = i;
	    } else if (controlBit4 == -1) {
		controlBit4 = i;
	    } else if (controlBit5 == -1) {
		controlBit5 = i;
	    } else if (controlBit6 == -1) {
		controlBit6 = i;
	    } else if (controlBit7 == -1) {
		controlBit7 = i;
	    } else if (controlBit8 == -1) {
		controlBit8 = i;
	    } else if (controlBit9 == -1) {
		controlBit9 = i;
	    }
	    counter++;
	}
    }

    if (counter > 9) {
	dieWithError("Too many control bits for NOT gate.");
    }

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if ((gateBitsArray[i] != BlankBit
	     && gateBitsArray[i] != ControlBit) &&
	    (i != target)) {
	    dieWithError("Unknown gate.");
	}
    }
    if (counter == 0) {
	// we've got a not gate
	getSymbolByIndex(target, symbolBuf);
	cout << "opX(" << symbolBuf << ");\n";
    } else if (counter == 1) {
	//we've got a CNot gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "cnot(" << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 2) {
	//we've got a toffoli (ccnot) gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "toffoli(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 3) {
	//we've got a c3not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c3not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 4) {
	//we've got a c4not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c4not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 5) {
	//we've got a c5not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c5not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit5, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 6) {
	//we've got a c6not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c6not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit5, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit6, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 7) {
	//we've got a c7not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c7not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit5, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit6, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit7, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 8) {
	//we've got a c8not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c8not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit5, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit6, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit7, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit8, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    } else if (counter == 9) {
	//we've got a c9not gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "c9not(" << symbolBuf << ", ";
	getSymbolByIndex(controlBit2, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit3, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit4, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit5, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit6, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit7, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit8, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(controlBit9, symbolBuf);
	cout << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    }
}

static void processHadamardGate(int target)
{
    char symbolBuf[50];
    int controlBit = -1;
    int counter=0;

    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] == ControlBit) {
	    controlBit = i;
	    counter++;
	}
    }

    if (counter > 1) {
	dieWithError("Too many control bits for hadmard gate.");
    }

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if ((gateBitsArray[i] != BlankBit
	     && gateBitsArray[i] != ControlBit) &&
	    (i != target)) {
	    dieWithError("Unknown gate.");
	}
    }
    if (counter == 0) {
	// we've got a regular hadamard gate
	getSymbolByIndex(target, symbolBuf);
	cout << "hadamard(" << symbolBuf << ");\n";
    } else if (counter == 1) {
	//we've got a controlled hadamard gate
	getSymbolByIndex(controlBit, symbolBuf);	    
	cout << "chadamard(" << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    }
}

static void processZGate(int target)
{
    char symbolBuf[50];
    int controlBit1 = -1;
    int controlBit2 = -1;
    int counter=0;

    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] == ControlBit) {
	    if (controlBit1 == -1) {
		controlBit1 = i;
	    } else {
		controlBit2 = i;
	    }
	    counter++;
	}
    }

    if (counter > 1) {
	dieWithError("Too many control bits for Z gate.");
    }

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if ((gateBitsArray[i] != BlankBit
	     && gateBitsArray[i] != ControlBit) &&
	    (i != target)) {
	    dieWithError("Unknown gate.");
	}
    }
    if (counter == 0) {
	// we've got a regualar Z gate
	getSymbolByIndex(target, symbolBuf);
	cout << "opZ(" << symbolBuf << ");\n";
    } else if (counter == 1) {
	//we've got a controlled Z gate
	getSymbolByIndex(controlBit1, symbolBuf);	    
	cout << "copZ(" << symbolBuf << ", ";
	getSymbolByIndex(target, symbolBuf);
	cout << symbolBuf << ");\n";
    }
}

static void processYGate(int target)
{
    char symbolBuf[50];

    // check for other junk in there
    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] != BlankBit && i != target) {
	    dieWithError("Unknown gate.");
	}
    }
    getSymbolByIndex(target, symbolBuf);
    cout << "opY(" << symbolBuf << ");\n";
}



static void analyzeGate()
{
    for(int i=0; i<numberOfVariables; i++) {
	if (gateBitsArray[i] == NotBit) {
	    processNotGate(i);
	    break;
	} else if (gateBitsArray[i] == MeasureBit) {
	    processMeasureGate(i);
	    break;
	} else if (gateBitsArray[i] == HadBit) {
	    processHadamardGate(i);
	    break;
	} else if (gateBitsArray[i] == YBit) {
	    processYGate(i);
	    break;
	} else if (gateBitsArray[i] == ZBit) {
	    processZGate(i);
	    break;
	} else if (gateBitsArray[i] == Reset1Bit) {
	    processReset1Gate(i);
	    break;
	} else if (gateBitsArray[i] == Reset0Bit) {
	    processReset0Gate(i);
	    break;
	} else if (gateBitsArray[i] == BlankBit) {
	    // continue
	}
    }
}

void processDeclaredGates()
{
    while (nextType == NameSymbType
	   && (strncmp(symbName, "qubit", 5) != 0)) {

	//SKIMP: DON'T KNOW WHAT TO DO WITH DECLARED GATES YET...
	//SKIMP: WILL JUST IGNORE THEM FOR NOW

	ignoreLine();
	getNextSymbol();

    }
}

int processQubitDecls()
{
    char varBuf[100];

    do {
	varBuf[0] = '\0';

	if (strncmp(symbName, "qubit", 5) != 0) {
	    dieWithError("Expected qubit declarations.");
	}
    
	getNextSymbol();
	expectPunct('=');
	expectPunct('"');

	do {
	    //assume declarations are only made of alphanumeric chars
	    if (nextType != NameSymbType && nextType != NumberSymbType) {
		dieWithError("Expecting qubit name.");
	    }

	    if (nextType == NumberSymbType) {
		sprintf(symbName, "_%i", (int)symbNumber);
	    }
		
	    strcat(varBuf, symbName); // keep on adding to end of existing
	                              // variable name string

	    getNextSymbol();
	} while (nextType != PuncSymbType || symbPunct != '"');

        if (varNameLookup(varBuf) != -1) {
            dieWithError("Duplicate variable name.");
        }
        if (isReservedWord(varBuf)) {
            dieWithError("Unexpected reserved word.");
        }

        // save name in symbol table
        if (! addToSymbolTab(varBuf, numberOfVariables)) {
            dieWithError("Cannot add to symbol table.");
        }
        numberOfVariables++;

	expectPunct('"');
    } while (nextType == NameSymbType && strncmp(symbName, "qubit", 5) == 0);

    return numberOfVariables;
}

void processGates()
{
    do {
	if (nextType != NameSymbType || strncmp(symbName, "gate", 4) != 0) {
	    dieWithError("Expected gate expression.");
	}

	getNextSymbol();
	expectPunct('=');
	expectPunct('{');

	for(int i=0; i<numberOfVariables-1; i++) {
	    parseBitOp(i); // parse that particular bit
	    getNextSymbol();
	    expectPunct(':');
	}
	parseBitOp(numberOfVariables-1);
	getNextSymbol();
	expectPunct('}');

	analyzeGate();
    } while (nextType == NameSymbType);
}

// end of $Source: /home/black/Quantum/QCSim/RCS/parseJAQ.C,v $
