// *created  "Tue Jun 10 14:19:28 2003" *by "Paul E. Black"
// *modified "Thu Mar 24 10:23:44 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	Grammar and functions to parse QHDL input
//-----------------------------------------------------------------------------

#include <complex.h>	// for complex amplitudes
#include <math.h>	// for pow()
#include <string.h>
#include "parsingFuncs.h"
#include "parseQHDL.h"

static string versionRCS = 
"$Header: /home/black/Quantum/QCSim/RCS/parseQHDL.C,v 1.14 2011/03/25 13:57:05 black Exp $";
string parseQHDL$version()
{
    return versionRCS;
}

//
//	VARIABLE DECLARATION STATEMENTS
//

extern int varNameLookup(char *name); // look up in symbol table
extern bool addToSymbolTab(char *name, int qubitNumb);

// reserved keywords
static const char *keywords[] = { "variable", "qubit" };
static int NUM_KEYWORDS = sizeof(keywords) / sizeof(*keywords);

static bool isReservedWord(char *buf)
{
    // check for reserved keywords
    for (int i = 0; i < NUM_KEYWORDS; i++) {
	if (strcmp(keywords[i], buf) == 0) {
	    return true;
	}
    }

    return false;		// if we get here everything is ok
}

static int numberOfVariables = 0;
static const int MAX_PARMS = 10;

// get one variable declaration statement
static void process_decl()
{
    // get variable names
    do {
	getNextSymbol(); // consume preceding variable keyword or ','

	if (nextType != NameSymbType) {
	    dieWithError("Expected variable name.");
	}
	if (varNameLookup(symbName) != -1) {
	    dieWithError("Duplicate variable name.");
	}
	if (isReservedWord(symbName)) {
	    dieWithError("Unexpected reserved word.");
	}

	// save name in symbol table
	if (! addToSymbolTab(symbName, numberOfVariables)) {
	    dieWithError("Cannot add to symbol table.");
	}
	numberOfVariables++;
	getNextSymbol(); // consume variable name
    } while (nextType == PuncSymbType && symbPunct == ',');

    expectPunct(':');

    // get type
    if (nextType != NameSymbType || strcmp(symbName, "qubit") != 0) {
	dieWithError("Keyword 'qubit' expected.");
    }
    getNextSymbol();

    expectPunct(';');
}

// get variable declaration statements
int get_decls()
{
    while (nextType == NameSymbType && strcmp(symbName, "variable") == 0) {
	process_decl();
    }

    return numberOfVariables;
}

//
//	(THE) INITIAL VALUE STATEMENT
//

// get a complex number: [+|-][float][i]
// nothing at all means 1
static float complex getAmplitude()
{
    float realAmp = 1;
    float complex amplitude;

    // look for leading + or -
    if (nextType == PuncSymbType) {
	if (symbPunct == '+') {
	    getNextSymbol(); // fine, already positive
	} else if (symbPunct == '-') {
	    realAmp = -1;
	    getNextSymbol();
	}
    }

    if (nextType == NumberSymbType) {
	realAmp *= symbNumber;
	getNextSymbol();
    }

    // now look for an i
    if (nextType == NameSymbType && strcmp(symbName, "i") == 0) {
	getNextSymbol();
	__real__ amplitude = 0;
	__imag__ amplitude = realAmp;
    } else {
	amplitude = realAmp; // imaginary part is 0
    }

    return amplitude;
}

// get a ket: |[01]+>.  return the index into the qubit array
static int getKet()
{
    int j = 0;			// counts number of bits in string
    int index_to_array = 0;	// index into 2^numqubits array

    expectPunct('|');

    if (nextType != NumberSymbType) {
	dieWithError("Bit string for ket expected.");
    }

    // check string and convert to an index
    while (symbName[j]) {
	index_to_array *= 2;
	if (symbName[j] == '1') {
	    index_to_array++;
	} else if (symbName[j] != '0') {
	    dieWithError("Ket must contain only 1's and 0's.");
	}
	j++;
    }
    if (j != numberOfVariables) {
	dieWithError("Number of bits in ket does not match number of variables.");
    }

    getNextSymbol(); // consume bit string

    expectPunct('>');

    return index_to_array;
}

static float normSquared(float complex c)
{
    return (__real__ c)*(__real__ c) + (__imag__ c)*(__imag__ c);
}

static void normalizeKet(int ketSize, float complex *ketp)
{
    float sum = 0;		// sum of norms of squares of amplitudes

    for (int i = 0; i < ketSize; i++) {
	// add up all the norms
	sum += (float) normSquared(ketp[i]);
    }

    if (sum <= 0) {
	dieWithError("Norm of ket amplitudes not positive.");
    }

    sum = (float) sqrt(sum);
    // normalize each entry
    for (int i = 0; i < ketSize; i++) {
	__real__ ketp[i] = __real__ ketp[i] / sum;
	__imag__ ketp[i] = __imag__ ketp[i] / sum;
    }
}

// process a constant ket, possibly a superposition
// put the (normalized) amplitudes in the initial ket
// If ketSize is 0, just read -- don't save (for qhdl2jaq)
static void processConstKet(int ketSize, float complex *ketp)
{
    // zero out amplitude array
    for (int i = 0; i < ketSize; i++) {
	ketp[i] = 0;
    }

    do {
	float complex amp = getAmplitude();
	int qindex = getKet();
	if (ketSize != 0) {
	    if (ketp[qindex] != 0.F) {
		// overwriting previously given ket
		dieWithError("Duplicate ket.");
	    }
	    ketp[qindex] = amp;
	}
    } while (nextType == PuncSymbType
			&& (symbPunct == '+' || symbPunct == '-'));

    if (ketSize != 0) {
	normalizeKet(ketSize, ketp);
    }
}

void get_init(int arraySize, float complex *initialKet)
{
    expectPunct('=');
    processConstKet(arraySize, initialKet);
    expectPunct(';');
}

//
//	GATES
//

static int processParms(int parraySize, paramType *parmArray)
{
    int numParams = 0;

    //first check for empty parameter list
    if (nextType == PuncSymbType && symbPunct == ')') {
        return 0;
    }

    while (nextType == NameSymbType || nextType == NumberSymbType
           || nextType == PuncSymbType) {
	if (numParams >= MAX_PARMS) {
	    dieWithError("Can't handle that many parameters.");
	}
        if (nextType == NameSymbType) {
            parmArray[numParams].i = varNameLookup(symbName);
            parmArray[numParams].type = 'q';
            if (parmArray[numParams].i < 0) {
                dieWithError("Invalid variable name.");
            }
        } else if (nextType == NumberSymbType) {
            parmArray[numParams].f = symbNumber;
            parmArray[numParams].type = 'n';
        } else if (nextType == PuncSymbType) {
	    float coeff = 1.0F; // whether a + or - in parm list

            if (symbPunct == '-') {
                coeff = -1.0F;
            } else if (symbPunct == '+') {
                //that's ok
            } else {
                dieWithError("Invalid parameter.");
            }
            getNextSymbol(); // get the float after the + or -
	    if (nextType != NumberSymbType) {
		dieWithError("Expecting floating-point number.");
	    }
            parmArray[numParams].f = coeff * symbNumber;
            parmArray[numParams].type = 'n';
        }

        getNextSymbol(); // consume parameter
        numParams++;

        // if there's no comma, exit loop.  Otherwise consume it.
        if (!(nextType == PuncSymbType && symbPunct == ',')) {
            break;
        }
        expectPunct(',');
    }
    return numParams;
}

int get_gate(int *funcIndexp, paramType *(parameters[]))
{
    extern int funcNameLookup(char *name);

    static paramType parmArray[MAX_PARMS];	// parameters passed to func

    // look up the function name in the table
    *funcIndexp = funcNameLookup(symbName);
    if (*funcIndexp < 0) {
	dieWithError("Unknown gate or command.");
    }
    getNextSymbol(); // consume gate name or command

    expectPunct('(');

    int numParms = processParms(MAX_PARMS, parmArray);

    expectPunct(')');
    expectPunct(';');

    *parameters = parmArray;
    return numParms;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/parseQHDL.C,v $
