/**************************************
 * Author:        Andrew Lane         *
 * Email:         alane@nist.gov      *
 * Date:          June 9, 2003        */
// *created  "Mon Jun  9 2003" *by "Andrew Lane"
// *modified "Thu Mar 24 10:25:03 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	Parsing functions
//-----------------------------------------------------------------------------

#include <ctype.h> // for isspace()
#include <stdlib.h> // for atof()
#include <stdio.h>
#include <string.h>
#include "parsingFuncs.h"

static string versionRCS = 
"$Header: /home/black/Quantum/QCSim/RCS/parsingFuncs.C,v 1.9 2011/03/25 13:57:26 black Exp $";
string parsingFuncs$version()
{
    return versionRCS;
}

// max length of a symbol, name, etc.
static const int MAX_SYMBOL_LEN = 100;
// column number in the file
static int fileCol = 1;
// line number in the file
static int fileLine = 1;
// line and column number in file of next character
static int nextChLine = 1, nextChCol = 1;

// the input stream
ifstream ins;

// SKIMP: THESE SHOULD BE EXTERNALLY READ-ONLY
// the type of symbol extracted from input stream
symbType nextType;
// next name, number, or punctuation
char symbName[MAX_SYMBOL_LEN+1]; // make sure it can be null delimited
float symbNumber;
char symbPunct;

//
//			      UTILITIES
//

// next character in the input stream
static char ch='\0';

void parmError(const char *gateName, int errNum)
{
    static const int bufsize = 100;
    char buf[bufsize];
    string errMsg = "Unknown error";
    switch (errNum) {
	case DupQubit:
	    errMsg = "Duplicate qubit not allowed";
	    break;
	case QubitOOR:
	    errMsg = "Qubit out of range";
	    break;
	case ProbOOR:
	    errMsg = "Probability out of range";
	    break;
	case ParmOOR:
	    errMsg = "Parameter out of range";
	    break;
    }
    (void)snprintf(buf, bufsize, "%s: %s.", gateName, errMsg.c_str());

    dieWithError(buf);
}

void dieWithError(const char *msgBuf)
{
    cout << "SYNTAX ERROR at line " << fileLine << 
		", column " << fileCol << ": [" <<
		msgBuf << "]\n";
    exit(1);
}

// Look for a particular punctuation.  If it is there, consume it.
// Otherwise, die with an error message.
void expectPunct(char pch)
{
    char messageBuf[20];

    if (nextType != PuncSymbType || symbPunct != pch) {
	sprintf(messageBuf, "'%c' expected.", pch);
	dieWithError(messageBuf);
    }
    getNextSymbol();
}

void nextTypeInfo()		// for debug
{
    if (nextType == NameSymbType) {
	cout << "next is name: " << symbName << endl;
    } else if (nextType == NumberSymbType) {
	cout << "next is number: " << symbNumber << endl;
    } else if (nextType == PuncSymbType) {
	cout << "next is punct: " << symbPunct << endl;
    } else {
	cout << "next is EOI" << endl;
    }
}

static bool getNextChar()
{
    // account for the character being consumed
    if (ch == '\n') {
	nextChLine++;
	nextChCol = 1;
    } else {
	// SKIMP: DOES NOT HANDLE TABS PROPERLY
	nextChCol++;
    }

    ins.get(ch);
    if (ins.fail()) {
	return false;
    }
//    cout << ch << " is at " << nextChLine << ", " << nextChCol << endl;
//    cout << "got [" << ch << "] from stream" << endl;

    return true;
}

// Skip leading comments and "white space"
// Put the next symbol in symbName, symbNumber, or symbPunct and
// the type in nextType.
void getNextSymbol()
{
    int i = 0;			// index of current letter in buffer

    // get first character to start things
    if (ch == '\0') {
	if (!getNextChar()) {
	    nextType = EOI;
	    return;
	}
    }

    // skip comments and leading "white space"
    while ((isspace(ch) || ch == '#')) {
	if (ch == '#') {
	    while (ch != '\n' && !ins.eof()) {
		getNextChar();
	    }
	}
	if (!getNextChar()) {
	    nextType = EOI;
	    return;
	}
    }

    if (ins.eof()) {		// no more input
	nextType = EOI;
	return;
    }

    // refer to where this symbol starts
    fileLine = nextChLine;
    fileCol  = nextChCol;

    if (isdigit(ch) || ch == '.') {
	// get unsigned number
	int decPointCount = 0;	// number of decimal points seen

	if (ch == '.') {
	    decPointCount++;
	}
	symbName[i++] = ch;
	getNextChar();
	while ((isdigit(ch) || ch == '.')) {
	    if (ch == '.') {
		decPointCount++;
	    }
	    // SKIMP: NO CHECK FOR BUFFER OVERFLOW
	    if (i >= MAX_SYMBOL_LEN) {
		dieWithError("Symbol too long.");
	    }
	    symbName[i++] = ch;
	    getNextChar();
	}
	symbName[i] = '\0';
	if (decPointCount > 1) {
	    dieWithError("Invalid number.");
	}
	symbNumber = atof(symbName);

	nextType = NumberSymbType;
    } else if (isalpha(ch) || ch == '_') {
	// get name
	symbName[i++] = ch;
	getNextChar();
	while ((isdigit(ch) || isalpha(ch)
		|| ch == '_')) {
	    if (i >= MAX_SYMBOL_LEN) {
		dieWithError("Symbol too long.");
	    }
	    symbName[i++] = ch;
	    getNextChar();
	}
	symbName[i] = '\0';
	nextType = NameSymbType;
    } else {
	// get one character
	symbPunct = ch;
	getNextChar();

	nextType = PuncSymbType;
    }
}

// ignore everything to the end of line
void ignoreLine()
{
    while (ch != '\n' && !ins.eof()) {
	getNextChar();
    }
}

// end of $Source: /home/black/Quantum/QCSim/RCS/parsingFuncs.C,v $
