/**************************************
 * Author:        Matthew Cass        *
 * Email:         mcass@nist.gov      *
 * Date:          June 6, 2002        */
// QHDL file parsing front end added by Andrew Lane
//	alane@nist.gov		May 30, 2003
// *modified "Thu Mar 24 10:37:41 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	Main file for the simulator to test aspects of quantum computing
//-----------------------------------------------------------------------------

#include <iostream>
#include <math.h>
#include <ctype.h>	// for toupper()
#include <stdlib.h>
#include <string.h>
#include "operations.h"
#include "parsingFuncs.h"
#include "parseQHDL.h"
#include "resource.h"

static std::string versionRCS =
"$Header: /home/black/Quantum/QCSim/RCS/qcsim.C,v 1.39 2011/03/25 13:59:55 black Exp $";

//-----------------------------------------------------------------------------
//	symbol table stuff
//-----------------------------------------------------------------------------

const int MAX_SYMBOLS = 50;	// max number of different symbols allowed

// information for one variable name
struct symbol_table_struct {
    char *name;		// name of symbol
    int qNo;		// qubit number, zero-based
};

symbol_table_struct symbol_table[MAX_SYMBOLS];

// number of variables in the symbol table
int symbolsInTable = 0;

// true if the symbol is added successfully
bool addToSymbolTab(char *name, int qubitNumb)
{
    // SKIMP: expand the table if we have no more room
    if (symbolsInTable >= MAX_SYMBOLS) return false;

    symbol_table[symbolsInTable].name = strdup(name);
    if (NULL == symbol_table[symbolsInTable].name) return false;
    symbol_table[symbolsInTable].qNo = qubitNumb;
    symbolsInTable++;

    return true;
}

// look up a name in the symbol table
int varNameLookup(char name[])
{
    for (int i = 0; i < symbolsInTable; i++) {
	if ((strcmp(name, symbol_table[i].name) == 0) &&
	    symbol_table[i].qNo != -1) {
	    return i;
	}
    }

    return -1;			// not found
}

// hard coded table of function names and information
static struct
{
    const char *funcName;		// name of function
    const char *parmTypes;		// types of parms: n(umber) or q(ubit)
} function_table[] = {
    {"cnot",		"qq"},
    {"copZ",		"qq"},
    {"hadamard", 	"q"},
    {"measure", 	"q"},
    {"noisyHadamard_01", "q"},
    {"traceOver", 	"q"},
    {"toffoli", 	"qqq"},
    {"opX",		"q"},
    {"opY",		"q"},
    {"opZ",		"q"},
    {"probXYZ", 	"nnnq"},
    {"chadamard", 	"qq"},
    {"print_state", 	""},
    {"genAmpDamp", 	"nnq"},
    {"BSWDepolCPhase",  "nqq"},
    {"rotateY", 	"nq"},
    {"c2not",		"qqq"},
    {"c3not",		"qqqq"},
    {"c4not",		"qqqqq"},
    {"c5not",		"qqqqqq"},
    {"c6not",		"qqqqqqq"},
    {"c7not",		"qqqqqqqq"},
    {"c8not",		"qqqqqqqqq"},
    {"c9not",		"qqqqqqqqqq"},
    {"c10not",		"qqqqqqqqqqq"},
};
int numFuncs = sizeof(function_table) / sizeof(*function_table);

static void performFunc(simstate *ssp, int ftindex, paramType *array)
{
    ParmErrorType rc=NoParmError;
    int tindex=-1;
    switch (ftindex) {
       case 0:// cnot
	rc = cnot(ssp, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);

	break;
      case 1:
	rc = copZ(ssp, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);

	break;
      case 2:
	rc = hadamard(ssp, symbol_table[array[0].i].qNo);
	break;
      case 3:
	rc = measure(ssp, symbol_table[array[0].i].qNo);
	break;
      case 4:
	rc = noisyHadamard_01(ssp, symbol_table[array[0].i].qNo);
	break;
      case 5:
	tindex = array[0].i;
	ssp->traceOver(symbol_table[tindex].qNo);

	// update symbol table after the qubit at "tindex" has been removed
	for (int i=0; i<symbolsInTable; i++) {
	    if (symbol_table[i].qNo > symbol_table[tindex].qNo) {
		symbol_table[i].qNo--;
	    }

	}
	symbol_table[tindex].qNo = -1;
	break;
      case 6:
      case 16: //c2not is alias for toffoli
	  rc = toffoli(ssp, symbol_table[array[2].i].qNo, symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo);

	break;
      case 7:
	// bit flip
	rc = opX(ssp, symbol_table[array[0].i].qNo);
	break;
      case 8:
	// pauli Y
	rc = opY(ssp, symbol_table[array[0].i].qNo);
	break;
      case 9:
	// pauli Z
	rc = opZ(ssp, symbol_table[array[0].i].qNo);
	break;
      case 10:
	rc = probabilisticXYZ(ssp, array[0].f, array[1].f, array[2].f, symbol_table[array[3].i].qNo);
	break;
      case 11:			// controlled hadamard
	rc = chadamard(ssp, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);
	break;

      case 12:
	cout << "current state: ";
	for(int i=0; i < symbolsInTable; i++) {
	    if (symbol_table[i].qNo != -1) { // if it's not been traced over
		cout << symbol_table[i].name;
		bool isLastQubitSymbol = true;
		for (int j=i+1; j < symbolsInTable; j++) {
		    if (symbol_table[j].qNo != -1) { // not been traced over
			isLastQubitSymbol = false;
			break;
		    }
		}		 
		if (!isLastQubitSymbol) {
		    cout << ", ";
		}
	    }
	}
	cout << endl;
	ssp->print_state();
	break;
      case 13:
	rc = genAmpDamp(ssp, array[0].f, array[1].f, symbol_table[array[2].i].qNo);
	break;
      case 14:
	rc = BSWDepolCPhase(ssp, array[0].f, symbol_table[array[1].i].qNo, symbol_table[array[2].i].qNo);
	break;
      case 15:
	rc = rotateY(ssp, array[0].f, symbol_table[array[1].i].qNo);
	break;
      case 17:
	  rc = c3not(ssp, symbol_table[array[3].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo);
	break;
      case 18:
	  rc = c4not(ssp, symbol_table[array[4].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo);
	break;
      case 19:
	  rc = c5not(ssp, symbol_table[array[5].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo);
	  break;
      case 20:
	  rc = c6not(ssp, symbol_table[array[6].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo);
	  break;
      case 21:
	  rc = c7not(ssp, symbol_table[array[7].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo);
	  break;
      case 22:
	  rc = c8not(ssp, symbol_table[array[8].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo,
		     symbol_table[array[7].i].qNo);
	  break;
      case 23:
	  rc = c9not(ssp, symbol_table[array[9].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo,
		     symbol_table[array[7].i].qNo,
		     symbol_table[array[8].i].qNo);
	  break;
      default:
	cout << "SYSTEM ERROR...INVALID FUNCTION CALL(" << ftindex << ")\n";
	exit(1);
    }
    if (rc != NoParmError) {
	parmError("Gate", rc);
    }
}

int funcNameLookup(char *name)
{
    for (int i = 0; i < numFuncs; i++) {
	if (strcmp(name, function_table[i].funcName) == 0) {
	    return i;
	}
    }
    return -1;
}

static void initialValue(simstate *ssp)
{
    int numQubits = ssp->Qubits();
    // size of the array of amplitudes - 2^n where n is the number of qubits
    int arraySize = (int) pow(2, numQubits);
    // array of amplitudes of the ket
    float complex *initialKet = new float complex [arraySize];
    saveMaxMemory();

    get_init(arraySize, initialKet);

    ssp->initialize(initialKet);

    delete [] initialKet;
}

static string useMsg = "qcsim [--version] file.qhdl";
static string qcsimVersion = "1.5";

int main(int argc, char *argv[])
{
    char *inputfile;
    int numQubits = 0;
    int reportResourcesUsed = 0;
    simstate *ssp;

    initialResources(); // save start time

    if (argc > 1 && strcmp(argv[1], "--version") == 0) {
	cout << "QCSim Version " << qcsimVersion << endl;
	cout << versionRCS << endl;
	cout << "using" << endl;
	cout << parseQHDL$version() << endl;
	cout << parsingFuncs$version() << endl;
	cout << operations$version() << endl;
	cout << simstate$version() << endl;
	cout << denmat$version() << endl;
	exit(0);
    }

    if (argc != 2) {
	cout << "Wrong number of operands" << endl;
	cout << "use: " << useMsg << endl;
	exit(-1);
    }

    inputfile = argv[1];

    ins.open(inputfile, ios::in);

    if (!ins) {
	cout << "Unable to open file: " << inputfile << endl;
	exit(-1);
    }

    cout << "Welcome to QCSim " << qcsimVersion << endl;

    // get the first symbol
    getNextSymbol();

    // get variable declaration statements
    numQubits = get_decls();

    if (numQubits < 1) {
	dieWithError("No variables declared.");
    }

    if (numQubits > 13) {
	cout << "Allocating more than 13 qubits hangs keystone" << endl;
	exit(0);
    }

    // allocate that big a state
    ssp = new simstate(numQubits);
    saveMaxMemory();

    // get (the) initialization statement and set to that ket
    initialValue(ssp);

    // get each gate and execute it
    while (nextType == NameSymbType) {
	int funcIndex;
	paramType *paramArray;
	int numParms = get_gate(&funcIndex, &paramArray);

	// check the number and types of parameters this function expects
	const char *parmTypes = function_table[funcIndex].parmTypes;
	int numParmsExpected = strlen(parmTypes);

	if (numParms != numParmsExpected) {
	    dieWithError("Wrong number of parameters for function.");
	}

	for (int pnum=0; pnum < numParms; pnum++) {
	    if (paramArray[pnum].type != parmTypes[pnum]) {
		dieWithError("Parameter type mismatch for function.");
	    }
	}

	performFunc(ssp, funcIndex, paramArray);
    }

    if (nextType != EOI) {
	// should be done with input now
	dieWithError("Unknown contents.");
    }

    ins.close();

    if (reportResourcesUsed) {
	// report time (and memory??) used
	reportResources();
    }

    return 0;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/qcsim.C,v $
