/**************************************
 * Author:        Andy Lane           *
 * Email:         alane@nist.gov      *
 * Date:          June 24, 2003       */
// *modified "Thu Mar 24 10:38:12 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	convert a QHDL file to jaQuzzi (.jaq) format
//-----------------------------------------------------------------------------

#include <iostream>
#include <math.h>
#include <ctype.h>	// for toupper()
#include <stdlib.h>
#include <string.h>
#include "operJaqOut.h"
#include "parsingFuncs.h"
#include "parseQHDL.h"

static string versionRCS =
"$Header: /home/black/Quantum/QCSim/RCS/qhdl2jaq.C,v 1.9 2011/03/25 14:00:02 black Exp $";

//-----------------------------------------------------------------------------
//	symbol table stuff
//-----------------------------------------------------------------------------

const int MAX_SYMBOLS = 50;	// max number of different symbols allowed

// information for one variable name
struct symbol_table_struct {
    char *name;		// name of symbol
    int qNo;		// qubit number, zero-based
};

symbol_table_struct symbol_table[MAX_SYMBOLS];

// number of variables in the symbol table
int symbolsInTable = 0;

// true if the symbol is added successfully
bool addToSymbolTab(char *name, int qubitNumb)
{
    // SKIMP: expand the table if we have no more room
    if (symbolsInTable >= MAX_SYMBOLS) return false;

    symbol_table[symbolsInTable].name = strdup(name);
    if (NULL == symbol_table[symbolsInTable].name) return false;
    symbol_table[symbolsInTable].qNo = qubitNumb;
    symbolsInTable++;

    return true;
}

// look up a name in the symbol table
int varNameLookup(char name[])
{
    for (int i = 0; i < symbolsInTable; i++) {
	if ((strcmp(name, symbol_table[i].name) == 0) &&
	    symbol_table[i].qNo != -1) {
	    return i;
	}
    }

    return -1;			// not found
}

// hard coded table of function names and information
static struct
{
    const char *funcName;		// name of function
    int numParms;		// number of parms this function takes
} function_table[] = {
    {"cnot", 2},
    {"copZ", 2},
    {"hadamard", 1},
    {"measure", 1},
    {"noisyHadamard_01", 1},
    {"traceOver", 1},
    {"toffoli", 3},
    {"opX", 1},
    {"opY", 1},
    {"opZ", 1},
    {"probXYZ", 4},
    {"chadamard", 2},
    {"print_state", 0},
    {"genAmpDamp", 3},
    {"BSWDepolCPhase", 3},
    {"rotateY", 2},
    {"c2not", 3},
    {"c3not", 4},
    {"c4not", 5},
    {"c5not", 6},
    {"c6not", 7},
    {"c7not", 8},
    {"c8not", 9},
    {"c9not", 10}
};
int numFuncs = sizeof(function_table) / sizeof(*function_table);

static void performFunc(int numQubits, int ftindex, paramType *array)
{
    ParmErrorType rc=NoParmError;

    switch (ftindex) {
      case 0:			// cnot
	rc = cnot(numQubits, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);

	break;
      case 1:
	rc = copZ(numQubits, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);

	break;
      case 2:
	rc = hadamard(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 3:
	rc = measure(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 4:
	rc = genericOneQubit(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 5:
	rc = genericOneQubit(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 6:
    case 16:
	rc = toffoli(numQubits, symbol_table[array[2].i].qNo, symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo);

	break;
      case 7:
	// bit flip
	rc = opX(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 8:
	// pauli Y
	rc = opY(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 9:
	// pauli Z
	rc = opZ(numQubits, symbol_table[array[0].i].qNo);
	break;
      case 10:
	rc = probabilisticXYZ(numQubits, array[0].f, array[1].f, array[2].f, symbol_table[array[3].i].qNo);
	break;
      case 11:
	rc = chadamard(numQubits, symbol_table[array[1].i].qNo, symbol_table[array[0].i].qNo);
	break;
      case 12:
	// do nothing for print state
	break;
    case 13:
        rc = genAmpDamp(numQubits, array[0].f, array[1].f, symbol_table[array[2].i].qNo);
        break;
      case 14:
	rc = BSWDepolCPhase(numQubits, array[0].f, symbol_table[array[1].i].qNo, symbol_table[array[2].i].qNo);
	break;
      case 15:
	rc = rotateY(numQubits, array[0].f, symbol_table[array[1].i].qNo);
	break;
      case 17:
	  rc = c3not(numQubits, symbol_table[array[3].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo);
	break;
      case 18:
	  rc = c4not(numQubits, symbol_table[array[4].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo);
	break;
      case 19:
	  rc = c5not(numQubits, symbol_table[array[5].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo);
	  break;
      case 20:
	  rc = c6not(numQubits, symbol_table[array[6].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo);
	  break;
      case 21:
	  rc = c7not(numQubits, symbol_table[array[7].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo);
	  break;
      case 22:
	  rc = c8not(numQubits, symbol_table[array[8].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo,
		     symbol_table[array[7].i].qNo);
	  break;
      case 23:
	  rc = c9not(numQubits, symbol_table[array[9].i].qNo,
		     symbol_table[array[0].i].qNo,
		     symbol_table[array[1].i].qNo,
		     symbol_table[array[2].i].qNo,
		     symbol_table[array[3].i].qNo,
		     symbol_table[array[4].i].qNo,
		     symbol_table[array[5].i].qNo,
		     symbol_table[array[6].i].qNo,
		     symbol_table[array[7].i].qNo,
		     symbol_table[array[8].i].qNo);
	  break;
      default:
	cout << "SYSTEM ERROR...INVALID FUNCTION CALL";
	exit(1);
    }
    if (rc != NoParmError) {
	parmError("Gate", rc);
    }
}

int funcNameLookup(char *name)
{
    for (int i = 0; i < numFuncs; i++) {
	if (strcmp(name, function_table[i].funcName) == 0) {
	    return i;
	}
    }
    return -1;
}

// read the initial value; don't do anything with it
static void initialValue()
{
    get_init(0, (float complex *)0);
}


static string useMsg = "qcsim [--version] file.qhdl";
static string qhdl2jaqVersion = "1.0";

int main(int argc, char *argv[])
{
    char *inputfile;
    int numQubits = 0;

    if (argc > 1 && strcmp(argv[1], "--version") == 0) {
	cout << "qhdl2jaq Version " << qhdl2jaqVersion << endl;
	cout << versionRCS << endl;
	cout << "using" << endl;
	cout << parseQHDL$version() << endl;
	cout << parsingFuncs$version() << endl;
	cout << operations$version() << endl;
	exit(0);
    }

    if (argc != 2) {
	cout << "Wrong number of operands" << endl;
	cout << "use: " << useMsg << endl;
	exit(-1);
    }

    inputfile = argv[1];

    ins.open(inputfile, ios::in);

    if (!ins) {
	cout << "Unable to open file: " << inputfile << endl;
	exit(-1);
    }

    //    cout << "Welcome to Qhdl2Jaq " << qhdl2jaqVersion << endl;

    // get the first symbol
    getNextSymbol();

    // get variable declaration statements
    numQubits = get_decls();

    for (int i=0; i<numQubits; i++) {
	cout << "qubit_" << i << "=\"" << symbol_table[i].name << "\"\n";
    }

    // read (the) initialization statement
    initialValue();

    // get each gate and execute it
    while (nextType == NameSymbType) {
	int funcIndex;
	paramType *paramArray;
	int numParms = get_gate(&funcIndex, &paramArray);

	// look up the function name in the table
	int numParmsExpected = function_table[funcIndex].numParms;

	if (numParms != numParmsExpected) {
	    dieWithError("Wrong number of parameters for function.");
	}

	performFunc(numQubits, funcIndex, paramArray);
    }

    if (nextType != EOI) {
	// should be done with input now
	dieWithError("Unknown contents.");
    }

    ins.close();

    return 0;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/qhdl2jaq.C,v $
