// *created  "Fri Sep  6 10:39:34 2002" *by "Paul E. Black"
// *modified "Wed Mar 23 18:08:41 2011" *by "Paul E. Black"

//-----------------------------------------------------------------------------
//
// This software was developed at the National Institute of Standards 
// and Technology by employees of the Federal Government in the course 
// of their official duties.  Pursuant to title 17 Section 105 of the 
// United States Code this software is not subject to copyright 
// protection and is in the public domain. 
// 
// We would appreciate acknowledgment if the software is used.
//
// Paul E. Black  paul.black@nist.gov
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// The state of the quantum computation at one time in a simulation.
// Currently it is a density matrix, but this could change.
//
//-----------------------------------------------------------------------------

#include "simstate.h"
#include "resource.h"
#include <math.h>	// for pow()

static std::string versionRCS =
"$Header: /home/black/Quantum/QCSim/RCS/simstate.C,v 1.6 2011/03/25 13:54:47 black Exp $";

std::string simstate$version() {
    return versionRCS;
}

// CONSTRUCTOR

simstate::simstate(int numQubits)
{
    mat = new denmat(numQubits);
    saveMaxMemory();
}

simstate::~simstate()
{
    delete mat;
}

void simstate::initialize(float complex initKet[])
{
    init_denmat(this->get_denmat(), initKet);
}

void simstate::traceOver(int qubit)
{
    if (qubit < 0 || qubit >= this->Qubits() ) {
	return;
    }

    // allocate a new density matrix with one fewer qubit
    denmat *newdmp = new denmat(this->Qubits() - 1);
    saveMaxMemory();

    // Consider the quadruple of density matrix entries with indices
    // having 0 and 1 at the targetBit, that is, indices of the form
    // p0q and p1q and r0s and r1s:
    //
    //	          p0q   p1q
    //      {      .     .     }
    //  r0s { ...  a ... b ... }
    //	    {      .     .     }
    //	    {      .     .     }
    //  r1s { ...  c ... d ... }
    //	    {      .     .     }
    //
    // The entry in the new, smaller density matrix is a+d or
    //    new_mat[rs][pq] = mat[r0s][p0q] + mat[r1s][p1q]
    // Abstractly we can just count through all indices pq and rs and
    // reshape each one to be p0q and p1q or r0s and r1s:
    //	for (int row = 0; row < size; row++) {
    //	    for (int col = 0; col < size; col++) {
    //		new_mat[row][col] = mat[r0s][p0q] + mat[r1s][p1q];
    //	    }
    //	}
    // But actually deriving r0s and r1s from row, etc. appears
    // bothersome, so we'll use the blocks-and-double-loop control 
    // structure from the OprOneQubit macro in opOneQubit.h.

    int tbit = (mat->bits()-1) - qubit;
    int blksz = (int)pow(2, tbit);
    int nextblock = 2 * blksz;
    int newrow = 0, newcol = 0;

    /* do every row of blocks */
    for (int rowblock = 0; rowblock < mat->size(); rowblock += nextblock) {
	/* do every row in that block */
	for (int row = rowblock; row < rowblock+blksz; row++) {
	    /* do every column of blocks */
	    for (int colblock = 0; colblock < mat->size();
						colblock += nextblock) {
		/* do every column in that block */
		for (int col = colblock; col < colblock+blksz; col++) {
/* uncomment this stuff to check array accesses
	be sure to #include <iostream>
		    if (row < 0 || row >= mat->size()) {
			cout << "invalid row: " << row << endl;
			return;
		    }
		    if (row+blksz < 0 || row+blksz >= mat->size()) {
			cout << "invalid row+blksz: " << row+blksz << endl;
			return;
		    }
		    if (col < 0 || col >= mat->size()) {
			cout << "invalid col: " << col << endl;
			return;
		    }
		    if (col+blksz < 0 || col+blksz >= mat->size()) {
			cout << "invalid col+blksz: " << col+blksz << endl;
			return;
		    }
		    if (newrow < 0 || newrow >= newdmp->size()) {
			cout << "invalid new row: " << newrow << endl;
			return;
		    }
		    if (newcol < 0 || newcol >= newdmp->size()) {
			cout << "invalid new col: " << newcol << endl;
			return;
		    }
 */
		    newdmp->set_state(newrow, newcol,
				      mat->get_state(row, col) +
				      mat->get_state(row+blksz, col+blksz));
		    newcol++;
		}
	    }
	    newcol = 0;
	    newrow++;
	}
    }

    // free old density matrix
    delete mat;

    // use new density matrix
    mat = newdmp;
}

// end of $Source: /home/black/Quantum/QCSim/RCS/simstate.C,v $
